#ifndef s_H_STATE
#define s_H_STATE
#include <vector>
#include <list>
#include "condition.hpp"

//All functions are inline, so here's a little summary of the class.
//class State {
//    private:
//       class Dummy_con
//       struct TriggerSet;
//
//       typedef std::vector< TriggerSet* > Inner;
//       typedef std::list< Inner > Outer;
//       std::list< std::vector< TriggerSet* > > outer;
//       std::list< std::vector< TriggerSet* > >::iterator cur_context;
//
//       int life;
//       void leave_context(const bool has_been_reset)
//       void use_triggerset(TriggerSet* t)
//    public:
//       enum{
//          START_OTHER = 1,
//          LEAVE_SELF = 2,
//          RESET_SELF = 4
//       };
//       bool is_active()
//       void enter()
//       bool accept(ALLEGRO_EVENT* e)
//       void add_condition(Condition c, int action, State* target)
//       State();
//       ~State();
// };

class State {
   private:
      class Dummy_con : public Condition {
         public:
            bool is_true(ALLEGRO_EVENT*){
               return false;
            }
            Condition* alias()
            {
               return this;
            }
            void unalias()
            {
               return;
            }
            Dummy_con()
               :Condition(0)
            {}
      };
      static Dummy_con dummy;

      struct TriggerSet{
         Condition* c;//Condition for trigger
         State* target;//Activate other state if not 0
         bool leave;//If true decrease life count: change to 0 state
         bool reset;//If true reset
         TriggerSet()
            : c(0), target(0), leave(false), reset(false)
         {}
         TriggerSet(TriggerSet& in)
            : c(in.c->alias()), target(in.target), leave(in.leave), reset(in.reset)
         {}
         ~TriggerSet() {
            c->unalias();
         }
      };

      //2D array: outer denotes instances of this state, inner contains TriggerSets of the states.
      typedef std::vector< TriggerSet* > Inner;
      typedef std::list< Inner > Outer;
      Outer outer;
      //The instance that is the current context
      Outer::iterator cur_context;

      int life;
      //Decrease life count and get rid of the current instance, if not the
      //last remaining (original instance). In the case of the original instance calls
      //Condition::reset() for all TriggerSets: reset is not normally called
      //for instances that get deleted.
      //Needs cur_context to be a dereferenceable iterator, but it increases it and
      //may set it to o.end()
      void leave_context(const bool has_been_reset){
         if (!life) return;
         Inner::iterator i = cur_context->begin();
         const Inner::iterator i_end = cur_context->end();
         if ( !(--life)) {
            if(!has_been_reset){
               for (; i != i_end; i++) {
                  (*i)->c->reset();
               }
            }
            cur_context++;//cur_context = outer.end()
         } else {
            //!! cur_context provided by accept(ALLEGRO_EVENT*): either
            //   outer.begin() or really the current context.
            for (; i != i_end; i++) {
               delete (*i);
            }
            cur_context = outer.erase(cur_context);
         }
      }
      //Needs cur_context to be a dereferenceable iterator, but it increases it and
      //may set it to o.end()
      void use_triggerset(TriggerSet* t) {
         if ( t->reset ){
            const Inner::iterator i_end = cur_context->end();
            for (Inner::iterator i = cur_context->begin(); i != i_end; i++) {
               (*i)->c->reset();
            }
         }
         if ( t->target ) t->target->enter();
         if ( t->leave ) leave_context(t->reset);//niet resetten, als dat al gebeurt is...;
            else cur_context++;
      }

   public:
      enum {
         START_OTHER = 1,
         LEAVE_SELF = 2,
         RESET_SELF = 4
      };
      bool is_active() const{
         return life;
      }
      void enter(){
         const Inner& original(outer.front());
         if (original.empty()) return;
         if (++life > 1) {
            outer.push_back(Inner());
            Inner& new_ts_list(outer.back());
            Inner::const_iterator end = original.end();
            for(Inner::const_iterator i = original.begin(); i != end; i++){
               new_ts_list.push_back(new TriggerSet(**i));
            }
         }
      }
      //Returns true if the event triggered something
      //Misschien: used, ignored, captured
      //!!Moet ook werken als hij niet actief is! ('parallel' met andere States die hem misschien uit zetten)
      bool accept(ALLEGRO_EVENT* e){
         bool used = false;
         Outer::iterator o_end = outer.end();
         //cur_context = outer.begin() -->last line of function
         while(cur_context != o_end) {
            Inner::iterator i_end = cur_context->end();
            for (Inner::iterator i = cur_context->begin(); i != i_end; i++) {
               if (e->type == (*i)->c->event_type && (*i)->c->is_true(e) ) {
                  use_triggerset(*i);
                  used = true;
                  break;//!! use_trigger changes current context
               }
            }
            if(!used) //!! use_trigger changes current context
               cur_context++;
         }
         cur_context = outer.begin();//!! See above in this function
         return used;
      }

      void add_condition(int action, Condition* c, State* target){ //Makes a TriggerSet (only if life=<1)
         if(life>1) return;//Imply life = 0
         TriggerSet* ts = new TriggerSet;
         if(action&START_OTHER && target != this)
            ts->target = target;
         ts->c = c;
         ts->leave = action&LEAVE_SELF;
         ts->reset = action&RESET_SELF;
         outer.front().push_back(ts);
      }

      State()
         : life(0)
      {
         outer.push_back(Inner());
         cur_context = outer.begin();
      }

      ~State(){
         while(life > 1)
            leave_context(true);//!! I dunno...
         Inner::iterator i_end = cur_context->end();
         for (Inner::iterator i = cur_context->begin(); i != i_end; i++) {
            delete (*i);
         }
      }
};
#endif